from ..type import ComponentType
from asn1PERser.classes.templates.creator import template_filler
from asn1PERser.classes.module import already_filled_template, typereference_to_type, constants


class SequenceType(ComponentType):
    def __init__(self):
        super(SequenceType, self).__init__()
        self._ComponentTypeList = []
        self.typereference = self.__class__.__name__

    def fill_template(self, has_parent=False):
        root_named_types = []
        addition_named_types = []
        subtypes_templates = []
        components_templates = []
        extension_addition_groups = []
        single_extension_group = []
        in_extension_addition_group = False
        extension_marker_present = False
        for ComponentType in self.ComponentTypeList:
            if ComponentType == '[[':
                in_extension_addition_group = True
                continue
            if ComponentType == ']]':
                in_extension_addition_group = False
                extension_addition_groups.append(list(single_extension_group))
                del single_extension_group[:]
                continue
            if in_extension_addition_group:
                single_extension_group.append(self._fill_named_types(ComponentType))
            if ComponentType == '...':
                extension_marker_present = True
                continue
            if not in_extension_addition_group:
                if not extension_marker_present:
                    root_named_types.append(self._fill_named_types(ComponentType))
                else:
                    addition_named_types.append(self._fill_named_types(ComponentType))
            if ComponentType.typereference in ['IntegerType', 'BooleanType', 'OctetStringType', 'BitStringType']:
                if ComponentType.is_constrained():
                    _, subtypes_template = ComponentType.fill_template(has_parent=True)
                    subtypes_templates.append(subtypes_template[:-2])    # remove 2 last new line characters
                continue
            elif ComponentType.typereference in ['ChoiceType', 'SequenceType', 'EnumeratedType', 'SequenceOfType']:
                inner_defined_templates, subtypes_template = ComponentType.fill_template(has_parent=True)
                subtypes_templates.append(subtypes_template[:-2])
                if inner_defined_templates:
                    components_templates.extend(inner_defined_templates)
                continue
            else:
                pass
            components_templates.append(ComponentType)
        filled_template = template_filler.fill(asn_type=self.__class__.__name__,
                                               class_name=self.template_class_name,
                                               class_type=self.__class__.__name__,
                                               subtemplates=subtypes_templates,
                                               extension_marker=extension_marker_present,
                                               root_named_types=root_named_types,
                                               addition_named_types=addition_named_types,
                                               extension_addition_groups=extension_addition_groups)
        if has_parent:
            return (components_templates, filled_template)
        if filled_template in already_filled_template:
            return ''
        already_filled_template.add(filled_template)
        filled_component_template_str = ''
        for component_template in components_templates:
            if component_template.default or component_template.typereference in typereference_to_type.keys(): # in constants.keys():
                continue
            filled_component_template_str += component_template.fill_template()
        return filled_component_template_str + filled_template

    @property
    def ComponentTypeList(self):
        return self._ComponentTypeList

    @ComponentTypeList.setter
    def ComponentTypeList(self, item):
        self._ComponentTypeList.append(item)

    def _fill_named_types(self, ComponentType):
        named_type_properties = {'optional': False,
                                 'default': False,
                                 'field_name': str(ComponentType.identifier),
                                 'field_type': str(ComponentType.template_field_type),
                                 'is_constant': False}
        if ComponentType.typereference in ['IntegerType', 'BooleanType', 'OctetStringType', 'BitStringType']:
            if ComponentType.is_constrained():
                named_type_properties['field_type'] = str(ComponentType.template_class_name)
        elif ComponentType.typereference in ['ChoiceType', 'SequenceType', 'EnumeratedType', 'SequenceOfType']:
            named_type_properties['field_type'] = str(ComponentType.template_class_name)
        if ComponentType.optional:
            named_type_properties['optional'] = True
        elif ComponentType.default is not None:
            named_type_properties['default'] = True
            if ComponentType.default in constants.keys():
                named_type_properties['is_constant'] = True;
                named_type_properties['default_value'] = str(ComponentType.default)
            else:
                if ComponentType.typereference in typereference_to_type.keys():
                    named_type_properties['default_value'] = typereference_to_type[ComponentType.typereference].parse_value(ComponentType.default)
                else:
                    named_type_properties['default_value'] = str(ComponentType.default)
        else:
            pass
        return named_type_properties

    def __getitem__(self, item):
        return self.ComponentTypeList[item]

    def __repr__(self):
        return '\n\t'.join([super(SequenceType, self).__repr__()] + \
                           ["ComponentTypes("] + \
                               [",".join([str(ComponentType) for ComponentType in self.ComponentTypeList])] + \
                           [")"])



import pytest
from pyasn1.type.namedval import NamedValues
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.classes.data.builtin.EnumeratedType import EnumeratedType
from asn1PERser.classes.types.constraint import ExtensionMarker


def SCHEMA_my_enum(enumerationRoot_list, extensionMarker_value=False):
    class MyEnum(EnumeratedType):
        '''
        MyEnum ::= ENUMERATED {
            e0,
            e1,
            .
            .
            .
            eN-1
            eN
        }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        enumerationRoot = NamedValues(
            *[(item, index) for index, item in enumerate(enumerationRoot_list)]
        )
        extensionAddition = NamedValues(
        )
        namedValues = enumerationRoot + extensionAddition
    return MyEnum


def SCHEMA_my_ext_enum(enumerationRoot_list, extensionAddition_list, extensionMarker_value=False):
    class MyEnum(EnumeratedType):
        '''
        MyEnum::= ENUMERATED
        {
            e0,
            e1,
            .
            .
            .
            eN - 1
            eN,
            ...,
            eN+1
            .
            .
            .
            eM-1,
            eM
        }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        enumerationRoot = NamedValues(
            *[(item, index) for index, item in enumerate(enumerationRoot_list)]
        )
        extensionAddition = NamedValues(
            *[(item, index) for index, item in enumerate(extensionAddition_list, start=len(enumerationRoot_list))]
        )
        namedValues = enumerationRoot + extensionAddition
    return MyEnum


def DATA_my_enum(enum, value):
    return enum(value)


short_enum = ['a0', 'a1']
enumeration_list = ['e0',   'e1',   'e2',   'e3',   'e4',   'e5',   'e6',   'e7',   'e8',   'e9',
                    'e10',  'e11',  'e12',  'e13',  'e14',  'e15',  'e16',  'e17',  'e18',  'e19',
                    'e20',  'e21',  'e22',  'e23',  'e24',  'e25',  'e26',  'e27',  'e28',  'e29',
                    'e30',  'e31',  'e32',  'e33',  'e34',  'e35',  'e36',  'e37',  'e38',  'e39',
                    'e40',  'e41',  'e42',  'e43',  'e44',  'e45',  'e46',  'e47',  'e48',  'e49',
                    'e50',  'e51',  'e52',  'e53',  'e54',  'e55',  'e56',  'e57',  'e58',  'e59',
                    'e60',  'e61',  'e62',  'e63',  'e64',  'e65',  'e66',  'e67',  'e68',  'e69',
                    'e70',  'e71',  'e72',  'e73',  'e74',  'e75',  'e76',  'e77',  'e78',  'e79',
                    'e80',  'e81',  'e82',  'e83',  'e84',  'e85',  'e86',  'e87',  'e88',  'e89',
                    'e90',  'e91',  'e92',  'e93',  'e94',  'e95',  'e96',  'e97',  'e98',  'e99',
                    'e100', 'e101', 'e102', 'e103', 'e104', 'e105', 'e106', 'e107', 'e108', 'e109',
                    'e110', 'e111', 'e112', 'e113', 'e114', 'e115', 'e116', 'e117', 'e118', 'e119',
                    'e120', 'e121', 'e122', 'e123', 'e124', 'e125', 'e126', 'e127', 'e128', 'e129',
                    'e130', 'e131', 'e132', 'e133', 'e134', 'e135', 'e136', 'e137', 'e138', 'e139',
                    'e140', 'e141', 'e142', 'e143', 'e144', 'e145', 'e146', 'e147', 'e148', 'e149',
                    'e150', 'e151', 'e152', 'e153', 'e154', 'e155', 'e156', 'e157', 'e158', 'e159',
                    'e160', 'e161', 'e162', 'e163', 'e164', 'e165', 'e166', 'e167', 'e168', 'e169',
                    'e170', 'e171', 'e172', 'e173', 'e174', 'e175', 'e176', 'e177', 'e178', 'e179',
                    'e180', 'e181', 'e182', 'e183', 'e184', 'e185', 'e186', 'e187', 'e188', 'e189',
                    'e190', 'e191', 'e192', 'e193', 'e194', 'e195', 'e196', 'e197', 'e198', 'e199',
                    'e200', 'e201', 'e202', 'e203', 'e204', 'e205', 'e206', 'e207', 'e208', 'e209',
                    'e210', 'e211', 'e212', 'e213', 'e214', 'e215', 'e216', 'e217', 'e218', 'e219',
                    'e220', 'e221', 'e222', 'e223', 'e224', 'e225', 'e226', 'e227', 'e228', 'e229',
                    'e230', 'e231', 'e232', 'e233', 'e234', 'e235', 'e236', 'e237', 'e238', 'e239',
                    'e240', 'e241', 'e242', 'e243', 'e244', 'e245', 'e246', 'e247', 'e248', 'e249',
                    'e250', 'e251', 'e252', 'e253', 'e254', 'e255', 'e256', 'e257', 'e258', 'e259']


@pytest.mark.parametrize("enumerated, encoded", [
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2]), 'e0'), '00'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2]), 'e1'), '80'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:10]), 'e9'), '90'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:17]), 'e9'), '48'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33]), 'e9'), '24'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33]), 'e32'), '80'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:100]), 'e98'), 'C4'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), 'e126'), '7E'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), 'e127'), '7F'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130]), 'e128'), '80'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e128'), '0080'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e254'), '00FE'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e255'), '00FF'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260]), 'e256'), '0100'),
])
def test_no_extension_marker_enumerated_can_be_encoded(enumerated, encoded):
    assert per_encoder(enumerated) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("enumerated, encoded", [
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2], extensionMarker_value=True), 'e0'), '00'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:2], extensionMarker_value=True), 'e1'), '40'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:10], extensionMarker_value=True), 'e9'), '48'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:17], extensionMarker_value=True), 'e9'), '24'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33], extensionMarker_value=True), 'e9'), '12'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:33], extensionMarker_value=True), 'e32'), '40'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:100], extensionMarker_value=True), 'e98'), '62'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), 'e126'), '3F00'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), 'e127'), '3F80'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:130], extensionMarker_value=True), 'e128'), '4000'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e128'), '000080'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e254'), '0000FE'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e255'), '0000FF'),
    (DATA_my_enum(SCHEMA_my_enum(enumerationRoot_list=enumeration_list[0:260], extensionMarker_value=True), 'e256'), '000100'),
])
def test_extension_marker_is_present_and_extension_addition_is_empty_but_value_is_from_root_can_be_encoded(enumerated, encoded):
    assert per_encoder(enumerated) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("enumerated, encoded", [
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:2], extensionAddition_list=short_enum, extensionMarker_value=True), 'e0'), '00'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:2], extensionAddition_list=short_enum, extensionMarker_value=True), 'e1'), '40'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:10], extensionAddition_list=short_enum, extensionMarker_value=True), 'e9'), '48'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:17], extensionAddition_list=short_enum, extensionMarker_value=True), 'e9'), '24'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:33], extensionAddition_list=short_enum, extensionMarker_value=True), 'e9'), '12'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:33], extensionAddition_list=short_enum, extensionMarker_value=True), 'e32'), '40'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:100], extensionAddition_list=short_enum, extensionMarker_value=True), 'e98'), '62'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum, extensionMarker_value=True), 'e126'), '3F00'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum, extensionMarker_value=True), 'e127'), '3F80'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:130], extensionAddition_list=short_enum, extensionMarker_value=True), 'e128'), '4000'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), 'e128'), '000080'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), 'e254'), '0000FE'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), 'e255'), '0000FF'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=enumeration_list[0:260], extensionAddition_list=short_enum, extensionMarker_value=True), 'e256'), '000100'),
])
def test_extension_marker_is_present_and_extension_addition_is_not_empty_but_value_is_from_root_can_be_encoded(enumerated, encoded):
    assert per_encoder(enumerated) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("enumerated, encoded", [
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:2], extensionMarker_value=True), 'e0'), '80'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:2], extensionMarker_value=True), 'e1'), '81'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:10], extensionMarker_value=True), 'e9'), '89'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:17], extensionMarker_value=True), 'e9'), '89'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:33], extensionMarker_value=True), 'e9'), '89'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:33], extensionMarker_value=True), 'e32'), 'A0'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:100], extensionMarker_value=True), 'e98'), 'C00162'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130], extensionMarker_value=True), 'e126'), 'C0017E'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130], extensionMarker_value=True), 'e127'), 'C0017F'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:130], extensionMarker_value=True), 'e128'), 'C00180'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'e128'), 'C00180'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'e254'), 'C001FE'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'e255'), 'C001FF'),
    (DATA_my_enum(SCHEMA_my_ext_enum(enumerationRoot_list=short_enum, extensionAddition_list=enumeration_list[0:260], extensionMarker_value=True), 'e256'), 'C0020100'),
])
def test_extension_marker_is_present_and_value_is_from_extension_can_be_encoded(enumerated, encoded):
    assert per_encoder(enumerated) == bytearray.fromhex(encoded)






import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.test.per.encoder.test_per_encode_bitstring import SCHEMA_constrained_bitstring, \
    SCHEMA_no_constrain_bitstring, SCHEMA_constrained_ext_bitstring, DATA_bitstring


@pytest.mark.parametrize("schema, encoded, value", [
    # (DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val=''), '00'),    # Does not work
    (SCHEMA_no_constrain_bitstring(), '0100', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='0')),
    (SCHEMA_no_constrain_bitstring(), '0180', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='1')),
    (SCHEMA_no_constrain_bitstring(), '0240', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='01')),
    (SCHEMA_no_constrain_bitstring(), '0280', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='10')),
    (SCHEMA_no_constrain_bitstring(), '02C0', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='11')),
    (SCHEMA_no_constrain_bitstring(), '0360', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='011')),
    (SCHEMA_no_constrain_bitstring(), '04A0', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='1010')),
    (SCHEMA_no_constrain_bitstring(), '04F0', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='1111')),
    (SCHEMA_no_constrain_bitstring(), '0FFFFE', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='111111111111111')),
    (SCHEMA_no_constrain_bitstring(), '18FFFFFF', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='111111111111111111111111')),
    (SCHEMA_no_constrain_bitstring(), '18800001', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='100000000000000000000001')),
    (SCHEMA_no_constrain_bitstring(), '18000001', DATA_bitstring(SCHEMA_no_constrain_bitstring(), bin_val='000000000000000000000001')),
    (SCHEMA_no_constrain_bitstring(), '2E0E0040FC0004', DATA_bitstring(SCHEMA_no_constrain_bitstring(),
                                                                       bin_val='0000111000000000010000001111110000000000000001')),
])
def test_no_constrain_bitstring_as_binary_value_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_no_constrain_bitstring(), '0400', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='0')),
    (SCHEMA_no_constrain_bitstring(), '0410', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='1')),
    (SCHEMA_no_constrain_bitstring(), '0810', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='10')),
    (SCHEMA_no_constrain_bitstring(), '0801', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='01')),
    (SCHEMA_no_constrain_bitstring(), '04A0', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='A')),
    (SCHEMA_no_constrain_bitstring(), '080A', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='0A')),
    (SCHEMA_no_constrain_bitstring(), '2C00000ABCDEF0', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='00000ABCDEF')),
    (SCHEMA_no_constrain_bitstring(), '30ABCDEFABCDEF', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF')),
    (SCHEMA_no_constrain_bitstring(), '34ABCDEFABCDEF00', DATA_bitstring(SCHEMA_no_constrain_bitstring(), hex_val='ABCDEFABCDEF0')),
    (SCHEMA_no_constrain_bitstring(), '58ABCDEFABCDEF0123456789', DATA_bitstring(SCHEMA_no_constrain_bitstring(),
                                                                                 hex_val='ABCDEFABCDEF0123456789')),
    (SCHEMA_no_constrain_bitstring(), '7CABCDEFABCDEF0123456789ABCDEFABC0', DATA_bitstring(SCHEMA_no_constrain_bitstring(),
                                                                                           hex_val='ABCDEFABCDEF0123456789ABCDEFABC')),
    (SCHEMA_no_constrain_bitstring(), '8080ABCDEFABCDEF0123456789ABCDEFABC0', DATA_bitstring(SCHEMA_no_constrain_bitstring(),
                                                                                             hex_val='ABCDEFABCDEF0123456789ABCDEFABC0')),
    (SCHEMA_no_constrain_bitstring(), '80B0ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789', DATA_bitstring(SCHEMA_no_constrain_bitstring(),
                                                                                                         hex_val='ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789')),

])
def test_no_constrain_bitstring_as_hex_value_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_bitstring(lb=0, ub=0), '00', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=0), bin_val='')),
])
def test_constrained_bitstring_of_zero_length_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_bitstring(lb=1, ub=1), '80', DATA_bitstring(SCHEMA_constrained_bitstring(lb=1, ub=1), bin_val='1')),
    (SCHEMA_constrained_bitstring(lb=1, ub=1), '00', DATA_bitstring(SCHEMA_constrained_bitstring(lb=1, ub=1), bin_val='0')),
    (SCHEMA_constrained_bitstring(lb=2, ub=2), '80', DATA_bitstring(SCHEMA_constrained_bitstring(lb=2, ub=2), bin_val='10')),
    (SCHEMA_constrained_bitstring(lb=2, ub=2), 'C0', DATA_bitstring(SCHEMA_constrained_bitstring(lb=2, ub=2), bin_val='11')),
    (SCHEMA_constrained_bitstring(lb=4, ub=4), 'A0', DATA_bitstring(SCHEMA_constrained_bitstring(lb=4, ub=4), bin_val='1010')),
    (SCHEMA_constrained_bitstring(lb=4, ub=4), '50', DATA_bitstring(SCHEMA_constrained_bitstring(lb=4, ub=4), bin_val='0101')),
    (SCHEMA_constrained_bitstring(lb=7, ub=7), 'AA', DATA_bitstring(SCHEMA_constrained_bitstring(lb=7, ub=7), bin_val='1010101')),
    (SCHEMA_constrained_bitstring(lb=7, ub=7), '54', DATA_bitstring(SCHEMA_constrained_bitstring(lb=7, ub=7), bin_val='0101010')),
    (SCHEMA_constrained_bitstring(lb=15, ub=15), 'CCCC', DATA_bitstring(SCHEMA_constrained_bitstring(lb=15, ub=15), bin_val='110011001100110')),
    (SCHEMA_constrained_bitstring(lb=15, ub=15), '3332', DATA_bitstring(SCHEMA_constrained_bitstring(lb=15, ub=15), bin_val='001100110011001')),
    (SCHEMA_constrained_bitstring(lb=16, ub=16), 'CCCC', DATA_bitstring(SCHEMA_constrained_bitstring(lb=16, ub=16), bin_val='1100110011001100')),
    (SCHEMA_constrained_bitstring(lb=16, ub=16), '3333', DATA_bitstring(SCHEMA_constrained_bitstring(lb=16, ub=16), bin_val='0011001100110011')),
    (SCHEMA_constrained_bitstring(lb=4, ub=4), 'A0', DATA_bitstring(SCHEMA_constrained_bitstring(lb=4, ub=4), hex_val='A')),
    (SCHEMA_constrained_bitstring(lb=8, ub=8), 'BB', DATA_bitstring(SCHEMA_constrained_bitstring(lb=8, ub=8), hex_val='BB')),
    (SCHEMA_constrained_bitstring(lb=12, ub=12), '1110', DATA_bitstring(SCHEMA_constrained_bitstring(lb=12, ub=12), hex_val='111')),
    (SCHEMA_constrained_bitstring(lb=16, ub=16), '2222', DATA_bitstring(SCHEMA_constrained_bitstring(lb=16, ub=16), hex_val='2222')),
])
def test_constrained_bitstring_same_length_gt_zero_and_le_16_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_bitstring(lb=17, ub=17), '333380', DATA_bitstring(SCHEMA_constrained_bitstring(lb=17, ub=17), bin_val='00110011001100111')),
    (SCHEMA_constrained_bitstring(lb=18, ub=18), '333340', DATA_bitstring(SCHEMA_constrained_bitstring(lb=18, ub=18), bin_val='001100110011001101')),
    (SCHEMA_constrained_bitstring(lb=19, ub=19), '333320', DATA_bitstring(SCHEMA_constrained_bitstring(lb=19, ub=19), bin_val='0011001100110011001')),
    (SCHEMA_constrained_bitstring(lb=20, ub=20), '333330', DATA_bitstring(SCHEMA_constrained_bitstring(lb=20, ub=20), bin_val='00110011001100110011')),
    (SCHEMA_constrained_bitstring(lb=21, ub=21), '333338', DATA_bitstring(SCHEMA_constrained_bitstring(lb=21, ub=21), bin_val='001100110011001100111')),
    (SCHEMA_constrained_bitstring(lb=22, ub=22), '333334', DATA_bitstring(SCHEMA_constrained_bitstring(lb=22, ub=22), bin_val='0011001100110011001101')),
    (SCHEMA_constrained_bitstring(lb=23, ub=23), '333332', DATA_bitstring(SCHEMA_constrained_bitstring(lb=23, ub=23), bin_val='00110011001100110011001')),
    (SCHEMA_constrained_bitstring(lb=24, ub=24), '333333', DATA_bitstring(SCHEMA_constrained_bitstring(lb=24, ub=24), bin_val='001100110011001100110011')),
    (SCHEMA_constrained_bitstring(lb=25, ub=25), '33333380', DATA_bitstring(SCHEMA_constrained_bitstring(lb=25, ub=25), bin_val='0011001100110011001100111')),
    (SCHEMA_constrained_bitstring(lb=128, ub=128), 'ABCDEFABCDEF0123456789ABCDEFABC0', DATA_bitstring(SCHEMA_constrained_bitstring(lb=128, ub=128),
                                                                                                      hex_val='ABCDEFABCDEF0123456789ABCDEFABC0')),
    (SCHEMA_constrained_bitstring(lb=176, ub=176), 'ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789', DATA_bitstring(SCHEMA_constrained_bitstring(lb=176, ub=176),
                                                                                                                  hex_val='ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789')),
])
def test_constrained_bitstring_same_length_gt_16_and_lt_64K_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_bitstring(lb=0, ub=1), '8080', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=1), bin_val='1')),
    (SCHEMA_constrained_bitstring(lb=0, ub=15), '1080', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=15), bin_val='1')),
    (SCHEMA_constrained_bitstring(lb=0, ub=15), 'F0E38E', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=15), bin_val='111000111000111')),
    (SCHEMA_constrained_bitstring(lb=14, ub=15), '80E38E', DATA_bitstring(SCHEMA_constrained_bitstring(lb=14, ub=15), bin_val='111000111000111')),
    (SCHEMA_constrained_bitstring(lb=0, ub=127), '1EE38E', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=127), bin_val='111000111000111')),
    (SCHEMA_constrained_bitstring(lb=0, ub=1000), '000FE38E', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=1000), bin_val='111000111000111')),
    (SCHEMA_constrained_bitstring(lb=0, ub=10000), '000FE38E', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=10000), bin_val='111000111000111')),
    (SCHEMA_constrained_bitstring(lb=0, ub=65535), '000FE38E', DATA_bitstring(SCHEMA_constrained_bitstring(lb=0, ub=65535), bin_val='111000111000111')),
])
def test_constrained_bitstring_with_ub_less_then_64K_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_ext_bitstring(lb=1, ub=1), '40', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=1, ub=1), bin_val='1')),
    (SCHEMA_constrained_ext_bitstring(lb=1, ub=1), '80077E', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=1, ub=1), bin_val='0111111')),
    (SCHEMA_constrained_ext_bitstring(lb=2, ub=2), '40', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=2, ub=2), bin_val='10')),
    (SCHEMA_constrained_ext_bitstring(lb=2, ub=2), '8003A0', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=2, ub=2), bin_val='101')),
    (SCHEMA_constrained_ext_bitstring(lb=7, ub=7), '800855', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=7, ub=7), bin_val='01010101')),
    (SCHEMA_constrained_ext_bitstring(lb=15, ub=15), '8011CCCD80', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=15, ub=15), bin_val='11001100110011011')),
    (SCHEMA_constrained_ext_bitstring(lb=16, ub=16), '8014333310', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=16, ub=16), bin_val='00110011001100110001')),
    (SCHEMA_constrained_ext_bitstring(lb=25, ub=25), '801C333333F0', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=25, ub=25), bin_val='0011001100110011001100111111')),
    (SCHEMA_constrained_ext_bitstring(lb=128, ub=128), '00ABCDEFABCDEF0123456789ABCDEFABC0', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=128, ub=128),
                                                                                                            hex_val='ABCDEFABCDEF0123456789ABCDEFABC0')),
    (SCHEMA_constrained_ext_bitstring(lb=176, ub=176), '00ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=176, ub=176),
                                                                                                                        hex_val='ABCDEFABCDEF0123456789ABCDEFABCDEF0123456789')),
    (SCHEMA_constrained_ext_bitstring(lb=0, ub=1), '4080', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=1), bin_val='1')),
    (SCHEMA_constrained_ext_bitstring(lb=0, ub=15), '0880', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=15), bin_val='1')),
    (SCHEMA_constrained_ext_bitstring(lb=0, ub=15), '78F0F0', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=15), bin_val='111100001111000')),
    (SCHEMA_constrained_ext_bitstring(lb=0, ub=128), '4000ABCDEFABCDEF0123456789ABCDEFABC0', DATA_bitstring(SCHEMA_constrained_ext_bitstring(lb=0, ub=128),
                                                                                                            hex_val='ABCDEFABCDEF0123456789ABCDEFABC0')),
])
def test_constrained_bitstring_with_extension_marker_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value



from pyparsing import Or, And, Forward, Literal, ZeroOrMore, OneOrMore, Optional, nestedExpr, Group
from asn1PERser.classes.parse_actions import parse_SequenceType, parse_ComponentType, \
    parse_ConstrainedType, parse_DateType, parse_TypeWithConstraint, \
    parse_RealType, parse_ChoiceType, parse_AlternativeTypeList, \
    parse_EnumeratedType, parse_OctetStringType, parse_BitStringType, parse_SequenceOfType, \
    parse_Constraint, parse_BooleanType, parse_IntegerType, parse_DefinedType, \
    parse_IA5String, parse_VisibleString, parse_NumericString, parse_UTF8String, parse_ExtensionAdditionAlternativesList
from .common_def import SignedNumber, NamedNumber, NamedNumberList, DefinedValue, NULL
from .lexical_items import identifier, LEFT_CURLY_BRACKET, RIGHT_CURLY_BRACKET, \
    COMMA, leftversionbrackets, rightversionbrackets, empty, \
    LEFT_PARENTHESIS, RIGHT_PARENTHESIS, EXCLAMATION_MARK, COLON, \
    CIRCUMFLEX_ACCENT, VERTICAL_LINE, LESS_THAN_SIGN, rangeseparator, \
    ellipsis, number, LEFT_SQUARE_BRACKET, RIGHT_SQUARE_BRACKET, \
    encodingreference, modulereference, typereference, FULL_STOP
from .value_def import Value


Type = Forward()
Constraint = nestedExpr(opener='(', closer=')').setParseAction(parse_Constraint)


BOOLEAN = Literal("BOOLEAN")
INTEGER = Literal("INTEGER")
OPTIONAL = Literal("OPTIONAL")
DEFAULT = Literal("DEFAULT")
COMPONENTS_OF = Literal("COMPONENTS OF")
BIT_STRING = Literal("BIT STRING")
ENUMERATED = Literal("ENUMERATED")
SEQUENCE = Literal("SEQUENCE")
SEQUENCE_OF = Literal("SEQUENCE OF")
CHOICE = Literal("CHOICE")
SET = Literal("SET")
OF = Literal("OF")
DATE = Literal("DATE")
IMPLICIT = Literal("IMPLICIT")
EXPLICIT = Literal("EXPLICIT")
UNIVERSAL = Literal("UNIVERSAL")
APPLICATION = Literal("APPLICATION")
PRIVATE = Literal("PRIVATE")

IA5String = Literal("IA5String").setParseAction(parse_IA5String)
VisibleString = Literal("VisibleString").setParseAction(parse_VisibleString)
NumericString = Literal("NumericString").setParseAction(parse_NumericString)
UTF8String = Literal("UTF8String").setParseAction(parse_UTF8String)



# 14.6
ExternalTypeReference = modulereference + FULL_STOP + typereference


# 14.1
DefinedType = Or([
    ExternalTypeReference,
    typereference,
    # ParameterizedType,         #  Rec. ITU-T X.683 | ISO/IEC 8824-4.
    # ParameterizedValueSetType  #  Rec. ITU-T X.683 | ISO/IEC 8824-4.
]).setParseAction(parse_DefinedType)


# 53.4
ExceptionIdentification = Or([
    SignedNumber,
    DefinedValue,
    Type + COLON + Value
])


# 53.4
ExceptionSpec = Or([
    EXCLAMATION_MARK + ExceptionIdentification,
    empty
])


# 21.1
RealType = Literal("REAL")
RealType.setParseAction(parse_RealType)


# 17.3
ReferencedType = Or([
    DefinedType,
    # UsefulType,
    # SelectionType,
    # TypeFromObject,        # ITU-T X.681 | ISO/IEC 8824-2, clause 15
    # ValueSetFromObjects    # ITU-T X.681 | ISO/IEC 8824-2, clause 15
])


# 17.5
NamedType = identifier + Type


# 18.1
BooleanType = BOOLEAN
BooleanType.setParseAction(parse_BooleanType)


# 19.1
IntegerType = Or([
    INTEGER,
    INTEGER + LEFT_CURLY_BRACKET + NamedNumberList + RIGHT_CURLY_BRACKET
]).setParseAction(parse_IntegerType)


# 20.1
EnumerationItem = Or([
    identifier,
    NamedNumber
])


# 20.1
Enumeration = Or([
    EnumerationItem + ZeroOrMore(COMMA + EnumerationItem)
    # EnumerationItem,    # ORIGINAL
    # EnumerationItem + COMMA + Enumeration    # ORIGINAL
])


# 20.1
AdditionalEnumeration = Enumeration


# 20.1
RootEnumeration = Enumeration


# 20.1
Enumerations = Or([
    RootEnumeration,
    RootEnumeration + COMMA + ellipsis + ExceptionSpec,
    RootEnumeration + COMMA + ellipsis + ExceptionSpec + COMMA + AdditionalEnumeration
])


# 20.1
EnumeratedType = (ENUMERATED + LEFT_CURLY_BRACKET +
                  Enumerations + RIGHT_CURLY_BRACKET).setParseAction(parse_EnumeratedType)


# 22.1
NamedBit = Or([
    identifier + LEFT_PARENTHESIS + number + RIGHT_PARENTHESIS,
    identifier + LEFT_PARENTHESIS + DefinedType + RIGHT_PARENTHESIS
])


# 22.1
NamedBitList = Or([
    NamedBit + ZeroOrMore(COMMA + NamedBit)
    # NamedBit,    # ORIGINAL
    # NamedBitList + COMMA + NamedBit    # ORIGINAL
])


# 22.1
BitStringType = Or([
    BIT_STRING,
    BIT_STRING + LEFT_CURLY_BRACKET + NamedBitList + RIGHT_CURLY_BRACKET
]).setParseAction(parse_BitStringType)


# 23.1
OctetStringType = Literal("OCTET STRING").setParseAction(parse_OctetStringType)


# 24.1
NullType = NULL


# 25.1
VersionNumber = Or([
    empty,
    number + COLON
])


# 25.1
ExtensionAndException = Or([
    ellipsis,
    ellipsis + ExceptionSpec
])


# 25.1
OptionalExtensionMarker = Or([
    COMMA + ellipsis,
    empty
])


# 25.1
ExtensionEndMarker = COMMA + ellipsis


# 25.1
ComponentType = Or([
    NamedType,
    NamedType + OPTIONAL,
    NamedType + DEFAULT + Value,
    # COMPONENTS_OF + Type
]).setParseAction(parse_ComponentType)


# 25.1
ComponentTypeList = Or([
    ComponentType + ZeroOrMore(COMMA + ComponentType)
    # ComponentType,
    # ComponentTypeList + COMMA + ComponentType,   #    Original from spec

])


# 25.1
RootComponentTypeList = ComponentTypeList


# 25.1
ExtensionAdditionGroup = And([
    leftversionbrackets,
    VersionNumber,
    ComponentTypeList,
    rightversionbrackets
])


# 25.1
ExtensionAddition = Or([
    ComponentType,
    ExtensionAdditionGroup
])


# 25.1
ExtensionAdditionList = Or([
    ExtensionAddition + ZeroOrMore(COMMA + ExtensionAddition),
    # ExtensionAddition,
    # ExtensionAdditionList + COMMA + ExtensionAddition    # ORIGINAL
])


# 25.1
ExtensionAdditions = Or([
    COMMA + ExtensionAdditionList,
    empty
])


# 25.1
ComponentTypeLists = Or([
    RootComponentTypeList,
    RootComponentTypeList + COMMA + ExtensionAndException +
        ExtensionAdditions + OptionalExtensionMarker,
    RootComponentTypeList + COMMA + ExtensionAndException +
        ExtensionAdditions + ExtensionEndMarker + COMMA +
        RootComponentTypeList,
    ExtensionAndException + ExtensionAdditions + ExtensionEndMarker +
        COMMA + RootComponentTypeList,
    ExtensionAndException + ExtensionAdditions + OptionalExtensionMarker
])


# 25.1
SequenceType = Or([
    SEQUENCE + LEFT_CURLY_BRACKET + RIGHT_CURLY_BRACKET,
    # SEQUENCE + LEFT_CURLY_BRACKET + ExtensionAndException +
    # OptionalExtensionMarker + RIGHT_CURLY_BRACKET,
    SEQUENCE + LEFT_CURLY_BRACKET + ComponentTypeLists + RIGHT_CURLY_BRACKET
]).setParseAction(parse_SequenceType)


# 26.1
SequenceOfType = Or([
    SEQUENCE_OF + Type,
    SEQUENCE_OF + NamedType
]).setParseAction(parse_SequenceOfType)


# 29.1
AlternativeTypeList = Or([
    NamedType + ZeroOrMore(COMMA + NamedType)
    # NamedType,    # ORIGINAL
    # AlternativeTypeList + COMMA + NamedType    # ORIGINAL
]).setParseAction(parse_AlternativeTypeList)


# This is not in ASN.1 spec but is needed to distinguish parsing of root vs extension types
ExtensionAdditionAlternativeTypeList = Or([
    NamedType + ZeroOrMore(COMMA + NamedType)
    # NamedType,    # ORIGINAL
    # AlternativeTypeList + COMMA + NamedType    # ORIGINAL
])


# 29.1
ExtensionAdditionAlternativesGroup = leftversionbrackets + VersionNumber + \
                                     ExtensionAdditionAlternativeTypeList + rightversionbrackets


# 29.1
ExtensionAdditionAlternative = Or([
    ExtensionAdditionAlternativesGroup,
    NamedType
]).setParseAction(parse_ExtensionAdditionAlternativesList)


# 29.1
ExtensionAdditionAlternativesList = Or([
    ExtensionAdditionAlternative + ZeroOrMore(COMMA + ExtensionAdditionAlternative)
    # ExtensionAdditionAlternative,    # ORIGINAL
    # ExtensionAdditionAlternativesList + COMMA + ExtensionAdditionAlternative    # ORIGINAL
])


# 29.1
ExtensionAdditionAlternatives = Or([
    COMMA + ExtensionAdditionAlternativesList,
    empty
])


# 29.1
RootAlternativeTypeList = AlternativeTypeList


# 29.1
AlternativeTypeLists = Or([
    RootAlternativeTypeList,
    RootAlternativeTypeList + COMMA + ExtensionAndException +
        ExtensionAdditionAlternatives + OptionalExtensionMarker
])


# 29.1
ChoiceType = (CHOICE + LEFT_CURLY_BRACKET +
              AlternativeTypeLists + RIGHT_CURLY_BRACKET).setParseAction(parse_ChoiceType)


# 38.4.1
DateType = DATE
DateType.setParseAction(parse_DateType)


# 31.2.1
Class = Or([
    UNIVERSAL,
    APPLICATION,
    PRIVATE,
    empty
])


# 31.2.1
ClassNumber = Or([
    number,
    DefinedValue
])


# 31.2.1
EncodingReference = Or([
    encodingreference + COLON,
    empty
])


# 31.2.1
Tag = LEFT_SQUARE_BRACKET + EncodingReference + Class + ClassNumber + RIGHT_SQUARE_BRACKET


# 31.2.1
TaggedType = Or([
    Tag + Type,
    Tag + IMPLICIT + Type,
    Tag + EXPLICIT + Type
])


# 31.1.5
PrefixedType = Or([
    TaggedType,
    # EncodingPrefixedType
])


# 41
RestrictedCharacterStringType = Or([
    # BMPString,
    # GeneralString,
    # GraphicString,
    IA5String,
    # ISO646String,
    NumericString,
    # PrintableString,
    # TeletexString,
    # T61String,
    # UniversalString,
    UTF8String,
    # VideotexString,
    VisibleString
])


# 40.1
CharacterStringType = Or([
    RestrictedCharacterStringType,
    # UnrestrictedCharacterStringType
])


# 17.2
BuiltinType = Or([
    BitStringType,
    BooleanType,
    CharacterStringType,
    ChoiceType,
    DateType,
    # DateTimeType,
    # DurationType ,
    # EmbeddedPDVType,
    EnumeratedType,
    # ExternalType,
    # InstanceOfType,
    IntegerType,
    # IRIType,
    NullType,
    # ObjectClassFieldType,
    # ObjectIdentifierType,
    OctetStringType,
    RealType,
    # RelativeIRIType,
    # RelativeOIDType,
    SequenceType,
    SequenceOfType,
    # SetType,
    # SetOfType,
    PrefixedType,
    # TimeType,
    # TimeOfDayType
])


# 49.5
TypeWithConstraint = Or([
    SET + Constraint + OF + Type,
    # SET + SizeConstraint + OF + Type,    # Present in original
    SEQUENCE + Constraint + OF + Type,
    # SEQUENCE + SizeConstraint + OF + Type,    # Present in original
    SET + Constraint + OF + NamedType,
    # SET + SizeConstraint + OF + NamedType,    # Present in original
    SEQUENCE + Constraint + OF + NamedType,
    # SEQUENCE + SizeConstraint + OF + NamedType    # Present in original
]).setParseAction(parse_TypeWithConstraint)


# 49.1
ConstrainedType = Or([
    BuiltinType + Constraint,
    TypeWithConstraint,
    # Type + Constraint,
    # Type + TypeWithConstraint    # ORIGINAL
]).setParseAction(parse_ConstrainedType)


# 17.1
Type << Or([
    Group(BuiltinType),
    Group(ReferencedType),
    Group(ConstrainedType)
])



import pytest
from pyasn1.type.namedtype import NamedType, NamedTypes, DefaultedNamedType, OptionalNamedType
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.classes.data.builtin.ChoiceType import ChoiceType
from asn1PERser.classes.data.builtin.IntegerType import IntegerType
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import ExtensionMarker


def SCHEMA_single_item_choice():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0  INTEGER
        }
        '''
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
        )
        componentType = rootComponent
    return MyChoice


def SCHEMA_no_extension_simple_choice():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0  INTEGER,
            i1  INTEGER
        }
        '''
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', IntegerType()),
        )
        componentType = rootComponent
    return MyChoice


def SCHEMA_reverse_no_extension_simple_choice():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i1  INTEGER,
            i0  INTEGER
        }
        '''
        rootComponent = AdditiveNamedTypes(
            NamedType('i1', IntegerType()),
            NamedType('i0', IntegerType()),
        )
        componentType = rootComponent
    return MyChoice


def SCHEMA_single_item_choice_with_extension():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0  INTEGER,
            ...
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
        )
        componentType = rootComponent
    return MyChoice


def SCHEMA_with_extension_simple_choice():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0  INTEGER,
            i1  INTEGER,
            ...
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', IntegerType()),
        )
        componentType = rootComponent
    return MyChoice


def SCHEMA_reverse_with_extension_simple_choice():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i1  INTEGER,
            i0  INTEGER,
            ...
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('i1', IntegerType()),
            NamedType('i0', IntegerType()),
        )
        componentType = rootComponent
    return MyChoice


def SCHEMA_selected_choice_from_extension_addition_one_elem():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0  INTEGER,
            ...
            i2  INTEGER,
            i3  INTEGER
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
        )
        extensionAddition = AdditiveNamedTypes(
            NamedType('i2', IntegerType()),
            NamedType('i3', IntegerType()),
        )
        componentType = rootComponent + extensionAddition
    return MyChoice


def SCHEMA_selected_choice_from_extension_addition_more_elems():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0    INTEGER,
            i1    INTEGER,
            i2    INTEGER,
            i3    INTEGER,
            i4    INTEGER,
            i5    INTEGER,
            i6    INTEGER,
            i7    INTEGER,
            i8    INTEGER,
            i9    INTEGER,
            i10   INTEGER,
            i11   INTEGER,
            i12   INTEGER,
            i13   INTEGER,
            i14   INTEGER,
            i15   INTEGER,
            i16   INTEGER,
            ...,
            i17   INTEGER,
            i18   INTEGER,
            i19   INTEGER,
            i20   INTEGER,
            i21   INTEGER,
            i22   INTEGER,
            i23   INTEGER,
            i24   INTEGER,
            i25   INTEGER,
            i26   INTEGER,
            i27   INTEGER,
            i28   INTEGER,
            i29   INTEGER,
            i30   INTEGER,
            i31   INTEGER,
            i32   INTEGER,
            i33   INTEGER,
            i34   INTEGER,
            i35   INTEGER
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', IntegerType()),
            NamedType('i2', IntegerType()),
            NamedType('i3', IntegerType()),
            NamedType('i4', IntegerType()),
            NamedType('i5', IntegerType()),
            NamedType('i6', IntegerType()),
            NamedType('i7', IntegerType()),
            NamedType('i8', IntegerType()),
            NamedType('i9', IntegerType()),
            NamedType('i10', IntegerType()),
            NamedType('i11', IntegerType()),
            NamedType('i12', IntegerType()),
            NamedType('i13', IntegerType()),
            NamedType('i14', IntegerType()),
            NamedType('i15', IntegerType()),
            NamedType('i16', IntegerType()),
        )
        extensionAddition = AdditiveNamedTypes(
            NamedType('i17', IntegerType()),
            NamedType('i18', IntegerType()),
            NamedType('i19', IntegerType()),
            NamedType('i20', IntegerType()),
            NamedType('i21', IntegerType()),
            NamedType('i22', IntegerType()),
            NamedType('i23', IntegerType()),
            NamedType('i24', IntegerType()),
            NamedType('i25', IntegerType()),
            NamedType('i26', IntegerType()),
            NamedType('i27', IntegerType()),
            NamedType('i28', IntegerType()),
            NamedType('i29', IntegerType()),
            NamedType('i30', IntegerType()),
            NamedType('i31', IntegerType()),
            NamedType('i32', IntegerType()),
            NamedType('i33', IntegerType()),
            NamedType('i34', IntegerType()),
            NamedType('i35', IntegerType()),
        )
        componentType = rootComponent + extensionAddition
    return MyChoice


def SCHEMA_selected_choice_from_extension_addition_groups_1():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0    INTEGER,
            i1    INTEGER,
            ...,
            i2    INTEGER,
            i3    INTEGER,
            i4    INTEGER,
            [[
            i5    INTEGER,
            i6    INTEGER
            ]],
            [[
            i7    INTEGER,
            i8    INTEGER,
            i9    INTEGER
            ]]
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', IntegerType()),
        )
        extensionAddition = AdditiveNamedTypes(
            NamedType('i2', IntegerType()),
            NamedType('i3', IntegerType()),
            NamedType('i4', IntegerType()),
        )
        extensionAdditionGroups = [
            AdditiveNamedTypes(
                NamedType('i5', IntegerType()),
                NamedType('i6', IntegerType()),
            ),
            AdditiveNamedTypes(
                NamedType('i7', IntegerType()),
                NamedType('i8', IntegerType()),
                NamedType('i9', IntegerType()),
            ),
        ]
        componentType = rootComponent + extensionAddition + extensionAdditionGroups
    return MyChoice


def SCHEMA_selected_choice_from_extension_addition_groups_2():
    class MyChoice(ChoiceType):
        '''
        MyChoice ::= CHOICE {
            i0    INTEGER,
            i1    INTEGER,
            ...,
            [[
            i5    INTEGER
            ]],
            [[
            i7    INTEGER,
            i8    INTEGER
            ]],
            [[
            i9    INTEGER
            ]]
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', IntegerType()),
        )
        extensionAdditionGroups = [
            AdditiveNamedTypes(
                NamedType('i5', IntegerType()),
            ),
            AdditiveNamedTypes(
                NamedType('i7', IntegerType()),
                NamedType('i8', IntegerType()),
            ),
            AdditiveNamedTypes(
                NamedType('i9', IntegerType()),
            ),
        ]
        componentType = rootComponent + extensionAdditionGroups
    return MyChoice



def DATA_choice(schema, identifier, value):
    myChoice = schema()
    myChoice[identifier] = value
    return myChoice


@pytest.mark.parametrize("choice, encoded", [
    (DATA_choice(SCHEMA_single_item_choice(), identifier='i0', value=10), '010A'),
    (DATA_choice(SCHEMA_single_item_choice(), identifier='i0', value=100000), '030186A0'),
    (DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i0', value=10), '00010A'),
    (DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i0', value=100000), '00030186A0'),
    (DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i1', value=10), '80010A'),
    (DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i1', value=100000), '80030186A0'),
    (DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i0', value=10), '80010A'),
    (DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i0', value=100000), '80030186A0'),
    (DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i1', value=10), '00010A'),
    (DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i1', value=100000), '00030186A0'),
])
def test_no_extension_marker_choice_can_be_encoded(choice, encoded):
    assert per_encoder(choice) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("choice, encoded", [
    (DATA_choice(SCHEMA_single_item_choice_with_extension(), identifier='i0', value=8), '000108'),
    (DATA_choice(SCHEMA_single_item_choice_with_extension(), identifier='i0', value=68496), '0003010B90'),
    (DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i0', value=-130), '0002FF7E'),
    (DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i0', value=23455), '00025B9F'),
    (DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i1', value=-1), '4001FF'),
    (DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i1', value=-100000), '4003FE7960'),
    (DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i0', value=10), '40010A'),
    (DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i0', value=100000), '40030186A0'),
    (DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i1', value=10), '00010A'),
    (DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i1', value=100000), '00030186A0'),
])
def test_only_extension_marker_preset_choice_can_be_encoded(choice, encoded):
    assert per_encoder(choice) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("choice, encoded", [
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i0', value=10), '00010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i1', value=10), '04010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i3', value=10), '0C010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i6', value=10), '18010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i9', value=10), '24010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i12', value=10), '30010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i15', value=10), '3C010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i16', value=10), '40010A'),
])
def test_extension_present_but_choice_from_root_can_be_encoded(choice, encoded):
    assert per_encoder(choice) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("choice, encoded", [
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i2', value=10), '8002010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i2', value=-2000), '800302F830'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i3', value=10), '8102010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i3', value=-2000), '810302F830'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i17', value=0), '80020100'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i17', value=10), '8002010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i18', value=-10), '810201F6'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i19', value=10), '8202010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i20', value=10), '8302010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i22', value=10), '8502010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i24', value=10), '8702010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i26', value=10), '8902010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i29', value=10), '8C02010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i33', value=10), '9002010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i34', value=10), '9102010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i35', value=10), '9202010A'),
])
def test_choice_from_extension_addition_can_be_encoded(choice, encoded):
    assert per_encoder(choice) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("choice, encoded", [
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i0', value=10), '00010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i1', value=10), '40010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i2', value=10), '8002010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i3', value=10), '8102010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i4', value=10), '8202010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i5', value=10), '8302010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i6', value=10), '8402010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i7', value=10), '8502010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i8', value=10), '8602010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i9', value=10), '8702010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i0', value=10), '00010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i1', value=10), '40010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i5', value=10), '8002010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i7', value=10), '8102010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i8', value=10), '8202010A'),
    (DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i9', value=10), '8302010A'),
])
def test_choice_from_extension_addition_groups_can_be_encoded(choice, encoded):
    assert per_encoder(choice) == bytearray.fromhex(encoded)



import os
import re
import logging
from asn1PERser.classes.templates import Template
from asn1PERser.classes.types.type import SimpleType, ComponentType, DefType


logger = logging.getLogger("asn1perser.moduleParser")

typereference_to_type = {}
valuereference_to_type = {}
constants = {}
already_filled_template = set()


class ModuleDefinition(Template):
    def __init__(self):
        self._ModuleIdentifier = ''
        self._EncodingReferenceDefault = ''
        self._TagDefault = ''
        self._ExtensionDefault = ''
        self._ModuleBody = ''

    def create_python_template(self, path=None):
        template = self.fill_template()
        if not path:
            return template
        with open(os.path.join(path, self.ModuleIdentifier + '.py'), 'w') as template_file:
            template_file.write(template)

    def fill_template(self):
        already_filled_template.clear()
        template = super(ModuleDefinition, self).fill_template()
        if self.ModuleBody:
            for Assignment in self.ModuleBody:
                template += Assignment.fill_template()
        return template

    def sort_assigment_list(self, AssignmentList):
        value_assignments = []
        constant_assignments = []
        simple_type_assignments = []
        component_type_assignments = []
        type_type_assignments = []
        for assignment in AssignmentList:
            if isinstance(assignment, SimpleType):
                if assignment.valuereference:
                    value_assignments.append(assignment)
                else:
                    simple_type_assignments.append(assignment)
            elif isinstance(assignment, DefType):
                if assignment.valuereference:
                    constant_assignments.append(assignment)
                    constants[assignment.valuereference] = assignment
                else:
                    type_type_assignments.append(assignment)
            else:
                component_type_assignments.append(assignment)
        component_type_assignments = self.sort_components(component_type_assignments)
        type_type_assignments, component_type_assignments = self.additional_sort(type_type_assignments, component_type_assignments)
        return [] + value_assignments + simple_type_assignments + type_type_assignments + constant_assignments + component_type_assignments

    def sort_components(self, component_type_assignments):
        MAX_ITERATIONS = 100
        all_sorted = False
        already_moved_typereferences = []
        new_order_typereferences = []
        new_order_components = []
        iter = 0

        while iter != MAX_ITERATIONS:
            components_typereferences = [component_type.typereference for component_type in component_type_assignments]
            if all_sorted:
                break
            all_sorted = True
            iter += 1
            for component_index in range(len(component_type_assignments)):
                ComponentType = component_type_assignments[component_index]
                for subcomponent in self._get_subcomponents(ComponentType):
                    if subcomponent.typereference in new_order_typereferences:
                        continue

                    try:
                        inner_component_index = components_typereferences.index(subcomponent.typereference)
                    except ValueError:
                        continue
                    if inner_component_index > component_index and subcomponent.typereference not in already_moved_typereferences:
                        new_order_typereferences.append(subcomponent.typereference)
                        new_order_components.append(component_type_assignments[inner_component_index])
                        already_moved_typereferences.append(subcomponent.typereference)
                        all_sorted = False
                else:
                    if ComponentType.typereference not in new_order_typereferences:
                        new_order_typereferences.append(ComponentType.typereference)
                        new_order_components.append(ComponentType)

            component_type_assignments = new_order_components[:]
            new_order_components = []
            new_order_typereferences = []
            already_moved_typereferences = []
        else:
            logger.warning("Could not sort components properly. This could lead to 'NameError' as some class"
                           " definition may be lower than its instantiation (Python reads .py files"
                           " from top to bottom.")
        return component_type_assignments

    def _get_subcomponents(self, component):
        subcomponents = []
        if component.__class__.__name__ == 'DefinedType':
            return [component]
        for subcomponent in component:
            if isinstance(subcomponent, str):  # elipsis: ...
                continue
            if subcomponent.typereference in ['IntegerType', 'BooleanType', 'OctetStringType',
                                              'BitStringType', 'EnumeratedType']:
                continue
            if subcomponent.typereference in ['ChoiceType', 'SequenceType', 'SequenceOfType']:
                subcomponents += self._get_subcomponents(subcomponent)
            else:
                subcomponents.append(subcomponent)
        return subcomponents

    def additional_sort(self, type_type_assignments, component_type_assignments):
        components_typereferences = [component_type.typereference for component_type in component_type_assignments]
        type_type_typereferences = [type_type.typereference for type_type in type_type_assignments]

        new_type_type_assignments = []

        for type_index in range(len(type_type_typereferences)):
            type_type_typereference = type_type_typereferences[type_index]
            try:
                type_in_components_index = components_typereferences.index(type_type_typereference)
            except ValueError:
                new_type_type_assignments.append(type_type_assignments[type_index])
                continue

            type_type_component = type_type_assignments[type_index]
            component_type_assignments.insert(type_in_components_index + 1, type_type_component)
            components_typereferences = [component_type.typereference for component_type in component_type_assignments]

        return (new_type_type_assignments, component_type_assignments)

    @property
    def ModuleIdentifier(self):
        return self._ModuleIdentifier

    @ModuleIdentifier.setter
    def ModuleIdentifier(self, ModuleIdentifier):
        self._ModuleIdentifier = ModuleIdentifier

    @property
    def EncodingReferenceDefault(self):
        return self._EncodingReferenceDefault

    @EncodingReferenceDefault.setter
    def EncodingReferenceDefault(self, EncodingReferenceDefault):
        self._EncodingReferenceDefault = EncodingReferenceDefault

    @property
    def TagDefault(self):
        return self._TagDefault

    @TagDefault.setter
    def TagDefault(self, TagDefault):
        self._TagDefault = TagDefault

    @property
    def ExtensionDefault(self):
        return self._ExtensionDefault

    @ExtensionDefault.setter
    def ExtensionDefault(self, ExtensionDefault):
        self._ExtensionDefault = ExtensionDefault

    @property
    def ModuleBody(self):
        return self._ModuleBody

    @ModuleBody.setter
    def ModuleBody(self, AssignmentList):
        self._ModuleBody = self.sort_assigment_list(AssignmentList)

    def __repr__(self):
        module_str = self.ModuleIdentifier
        if self.EncodingReferenceDefault:
            module_str += ' ' + self.EncodingReferenceDefault
        if self.TagDefault:
            module_str += ' ' + self.TagDefault
        if self.ExtensionDefault:
            module_str += ' ' + self.ExtensionDefault
        module_str += ' ' + 'BEGIN' + '\n'
        if self.ModuleBody:
            for Assignment in self.ModuleBody:
                module_str += str(Assignment)
        module_str += 'END'
        return module_str


def remove_comments(asn1_schema):
    return _remove_line_comments(asn1_schema)


def _remove_line_comments(asn1_schema):
    no_comments = ''
    for line in asn1_schema.splitlines():
        line = re.sub('--.*', '', line)
        line += '\n'
        no_comments += line
    return no_comments



from pyasn1.type.univ import Boolean
from asn1PERser.codec.per.encoder import encode_boolean
from asn1PERser.codec.per.decoder import decode_boolean


class BooleanType(Boolean):
    def fill_field_list(self, field_list):
        boolean_field_list = encode_boolean(self)
        if boolean_field_list:
            field_list.extend(boolean_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_boolean(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        return bool(self)



from ..type import SimpleType
from asn1PERser.classes.templates.creator import template_filler
from asn1PERser.classes.module import already_filled_template


class OctetStringType(SimpleType):
    def __init__(self):
        super(OctetStringType, self).__init__()
        self.typereference = self.__class__.__name__

    def fill_template(self, has_parent=False):
        if self.Value:
            return template_filler.fill(asn_type=self.__class__.__name__, type_or_value='value',
                                        valuereference=self.valuereference, value=self.Value,
                                        class_type=self.template_field_type)
        filled_template = template_filler.fill(asn_type=self.__class__.__name__, type_or_value='type',
                                               class_name=self.template_class_name, class_type=self.__class__.__name__,
                                               constraint={'extensionMarker': self.constraint.extensionMarker,
                                                           'lowerEndpoint': self.constraint.lowerEndpoint,
                                                           'upperEndpoint': self.constraint.upperEndpoint})
        if has_parent:
            return (None, filled_template)
        if filled_template not in already_filled_template:
            already_filled_template.add(filled_template)
            return filled_template
        return ''

    def parse_value(self, value):
        if value.endswith('B'):
            return "binValue={}".format(value[:-1])
        if value.endswith('H'):
            return "hexValue={}".format(value[:-1])



import pytest
from pyasn1.type.namedtype import NamedType, NamedTypes, DefaultedNamedType, OptionalNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.codec.per.encoder import MandatoryExtensionFieldNotPresent
from asn1PERser.classes.data.builtin.SequenceType import SequenceType
from asn1PERser.classes.data.builtin.IntegerType import IntegerType
from asn1PERser.classes.data.builtin.BooleanType import BooleanType
from asn1PERser.classes.data.builtin.OctetStringType import OctetStringType
from asn1PERser.classes.data.builtin.BitStringType import BitStringType
from asn1PERser.classes.data.builtin.EnumeratedType import EnumeratedType
from asn1PERser.classes.types.constraint import ValueRange, ValueSize, ExtensionMarker


I0_VAL = 34567
I1_VAL = -845
I2_VAL = 23
I3_VAL = 200000
I4_VAL = 5


class i0(IntegerType):
    pass


class i1(IntegerType):
    pass


class i2(IntegerType):
    subtypeSpec = ValueRange(0, 100000, extensionMarker=True)


class i3(IntegerType):
    subtypeSpec = ValueRange(128, 1000, extensionMarker=True)


class i4(IntegerType):
    subtypeSpec = ValueRange(0, 10, extensionMarker=True)


class o2(OctetStringType):
    subtypeSpec = ValueSize(2, 6)


class b0(BitStringType):
    subtypeSpec = ValueSize(0, 10, extensionMarker=True)


class b4(BitStringType):
    subtypeSpec = ValueSize(5, 5)


class b6(BitStringType):
    subtypeSpec = ValueSize(20, 1000)


class b9(BitStringType):
    subtypeSpec = ValueSize(1, 2)


class b10(BitStringType):
    subtypeSpec = ValueSize(20, 20)


class e0(EnumeratedType):
    subtypeSpec = ExtensionMarker(True)
    enumerationRoot = NamedValues(
        ('a0', 0),
        ('a1', 1)
    )
    namedValues = enumerationRoot


class e1(EnumeratedType):
    subtypeSpec = ExtensionMarker(True)
    enumerationRoot = NamedValues(
        ('b0', 0)
    )
    extensionAddition = NamedValues(
        ('b1', 1),
        ('b2', 2)
    )
    namedValues = enumerationRoot + extensionAddition


class e2(EnumeratedType):
    enumerationRoot = NamedValues(
        ('c0', 0),
        ('c1', 1),
        ('c2', 2)
    )
    namedValues = enumerationRoot


class e3(EnumeratedType):
    subtypeSpec = ExtensionMarker(True)
    enumerationRoot = NamedValues(
        ('d0', 0),
        ('d1', 1),
        ('d2', 2)
    )
    namedValues = enumerationRoot


class e4(EnumeratedType):
    subtypeSpec = ExtensionMarker(True)
    enumerationRoot = NamedValues(
        ('f0', 0)
    )
    extensionAddition = NamedValues(
        ('f1', 1)
    )
    namedValues = enumerationRoot + extensionAddition


class e5(EnumeratedType):
    enumerationRoot = NamedValues(
        ('g0', 0)
    )
    namedValues = enumerationRoot


def SCHEMA_two_elem_seq_with_extension_marker_present_or_not(extensionMarker_value):
    class MySeq(SequenceType):
        '''
        If extensionMarker_value == False:
            MySeq ::= SEQUENCE
            {
             i0    INTEGER,
             i1    INTEGER
            }

        else:
            MySeq ::= SEQUENCE
            {
             i0    INTEGER,
             i1    INTEGER
             ...
            }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        rootComponent = NamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', IntegerType())
        )
        componentType = rootComponent
    return MySeq()


def SCHEMA_seq_with_only_extension_marker_present_or_not(extensionMarker_value):
    class MySeq(SequenceType):
        '''
        If extensionMarker_value == False:
            MySeq ::= SEQUENCE
            {
             i0    INTEGER,
             i1    INTEGER,
             i2    INTEGER (0..100000),
             i3    INTEGER (128..1000,...),
             i4    INTEGER (0..10,...)
            }

        else:
            MySeq ::= SEQUENCE
            {
             i0    INTEGER,
             i1    INTEGER,
             i2    INTEGER (0..100000),
             i3    INTEGER (128..1000,...),
             i4    INTEGER (0..10,...),
             ...
            }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        rootComponent = NamedTypes(
            NamedType('i0', IntegerType()),
            NamedType('i1', IntegerType()),
            NamedType('i2', i2()),
            NamedType('i3', i3()),
            NamedType('i4', i4()),
        )
        componentType = rootComponent
    return MySeq()


def SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value):
    class MySeq(SequenceType):
        '''
            MySeq ::= SEQUENCE
            {
             i0    INTEGER,
             ...,
             i1    INTEGER,
             i2    INTEGER (0..100000),
             i3    INTEGER (128..1000,...),
             i4    INTEGER (0..10,...)
            }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType())
        )
        extensionAddition = AdditiveNamedTypes(
            NamedType('i1', IntegerType()),
            NamedType('i2', i2()),
            NamedType('i3', i3()),
            NamedType('i4', i4())
        )
        componentType = rootComponent + extensionAddition
    return MySeq()


def SCHEMA_seq_with_some_optional_elements_in_root(extensionMarker_value):
    class SeqOptional(SequenceType):
        '''
        MySeq ::= SEQUENCE
        {
            i0    INTEGER,
            i1    INTEGER OPTIONAL,
            i2    INTEGER (0..100000),
            i3    INTEGER (128..1000,...) OPTIONAL,
            i4    INTEGER (0..10,...)
        }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            OptionalNamedType('i1', IntegerType()),
            NamedType('i2', i2()),
            OptionalNamedType('i3', i3()),
            NamedType('i4', i4())
        )
        componentType = rootComponent
    return SeqOptional()


def SCHEMA_seq_with_only_optional_elements_in_root(extensionMarker_value):
    class SeqOptional(SequenceType):
        '''
        MySeq ::= SEQUENCE
        {
            i0  [0]  INTEGER OPTIONAL,
            i1       INTEGER OPTIONAL
        }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        rootComponent = AdditiveNamedTypes(
            OptionalNamedType('i0', IntegerType()),
            OptionalNamedType('i1', IntegerType())
        )
        componentType = rootComponent
    return SeqOptional()


def SCHEMA_seq_with_default_elements_in_root(extensionMarker_value):
    class SeqOptional(SequenceType):
        '''
        MySeq ::= SEQUENCE
        {
            i0    INTEGER,
            i1    INTEGER DEFAULT 10,
            i2    INTEGER (0..100000),
            i3    INTEGER (128..1000,...) DEFAULT 300,
            i4    INTEGER (0..10,...)
        }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        rootComponent = AdditiveNamedTypes(
            NamedType('i0', IntegerType()),
            DefaultedNamedType('i1', IntegerType(10)),
            NamedType('i2', i2()),
            DefaultedNamedType('i3', i3(300)),
            NamedType('i4', i4())
        )
        componentType = rootComponent
    return SeqOptional()


def SCHEMA_sequence_of_boolean(extensionMarker_value=True):
    class SeqOfBool(SequenceType):
        '''
        MySeq ::= SEQUENCE
        {
            b0    BOOLEAN,
            b1    BOOLEAN,
            b2    BOOLEAN,
            b3    BOOLEAN,
            b4    BOOLEAN,
            b5    BOOLEAN,
            b6    BOOLEAN,
            b7    BOOLEAN,
            b8    BOOLEAN,
            b9    BOOLEAN,
            ...,
            b10   BOOLEAN,
            b11   BOOLEAN,
            b12   BOOLEAN,
            b13   BOOLEAN,
            b14   BOOLEAN,
            b15   BOOLEAN,
            b16   BOOLEAN,
            b17   BOOLEAN,
            b18   BOOLEAN,
            b19   BOOLEAN,
            b20   BOOLEAN
        }
        '''
        subtypeSpec = ExtensionMarker(extensionMarker_value)
        rootComponent = AdditiveNamedTypes(
            NamedType('b0', BooleanType()),
            NamedType('b1', BooleanType()),
            NamedType('b2', BooleanType()),
            NamedType('b3', BooleanType()),
            NamedType('b4', BooleanType()),
            NamedType('b5', BooleanType()),
            NamedType('b6', BooleanType()),
            NamedType('b7', BooleanType()),
            NamedType('b8', BooleanType()),
            NamedType('b9', BooleanType()),
        )
        extensionAddition = AdditiveNamedTypes(
            NamedType('b10', BooleanType()),
            NamedType('b11', BooleanType()),
            NamedType('b12', BooleanType()),
            NamedType('b13', BooleanType()),
            NamedType('b14', BooleanType()),
            NamedType('b15', BooleanType()),
            NamedType('b16', BooleanType()),
            NamedType('b17', BooleanType()),
            NamedType('b18', BooleanType()),
            NamedType('b19', BooleanType()),
            NamedType('b20', BooleanType()),
        )
        componentType = rootComponent + extensionAddition
    return SeqOfBool()


def SCHEMA_sequence_of_octetstring():
    class SeqOfOctet(SequenceType):
        '''
        MySeq ::= SEQUENCE {
            o0    OCTET STRING,
            o1    OCTET STRING OPTIONAL,
            o2    OCTET STRIMG (SIZE(2..6)),
            o3    OCTET STRING,
            o4    OCTET STRING DEFAULT 'FFFFFF'H,
            o5    OCTET STRING OPTIONAL,
            o6    OCTET STRING
        }
        '''
        rootComponent = AdditiveNamedTypes(
            NamedType('o0', OctetStringType()),
            OptionalNamedType('o1', OctetStringType()),
            NamedType('o2', o2()),
            NamedType('o3', OctetStringType()),
            DefaultedNamedType('o4', OctetStringType(hexValue='FFFFFF')),
            OptionalNamedType('o5', OctetStringType()),
            NamedType('o6', OctetStringType()),
        )
        componentType = rootComponent
    return SeqOfOctet()


def SCHEMA_sequence_of_bitstring():
    class SeqOfBit(SequenceType):
        '''
        MySeq  ::= SEQUENCE {
            b0    BIT STRING (SIZE(0..10, ...)),
            b1    BIT STRING,
            b2    BIT STRING DEFAULT 'AAAA'H,
            b3    BIT STRING OPTIONAL,
            b4    BIT STRING (SIZE(5..5)),
            b5    BIT STRING,
            b6    BIT STRING (SIZE(20..1000)),
            b7    BIT STRING,
            b8    BIT STRING,
            ...,
            b9    BIT STRING (SIZE(1..2)) OPTIONAL,
            b10   BIT STRING (SIZE(20..20)),
            b11   BIT STRING
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('b0', b0()),
            NamedType('b1', BitStringType()),
            DefaultedNamedType('b2', BitStringType(hexValue='AAAA')),
            OptionalNamedType('b3', BitStringType()),
            NamedType('b4', b4()),
            NamedType('b5', BitStringType()),
            NamedType('b6', b6()),
            NamedType('b7', BitStringType()),
            NamedType('b8', BitStringType())
        )
        extensionAddition = AdditiveNamedTypes(
            OptionalNamedType('b9', b9()),
            NamedType('b10', b10()),
            NamedType('b11', BitStringType())
        )
        componentType = rootComponent + extensionAddition
    return SeqOfBit()


def SCHEMA_sequence_of_enumerated():
    class SeqOfEnum(SequenceType):
        '''
        SeqOfEnum ::= SEQUENCE {
            e0    ENUMERATED {a0, a1, ...},
            e1    ENUMERATED {b0, ..., b1, b2} OPTIONAL,
            e2    ENUMERATED {c0, c1, c2},
            e3    ENUMERATED {d0, d1, d2, ...} DEFAULT d1,
            ...,
            e4    ENUMERATED {f0, ..., f1} OPTIONAL,
            e5    ENUMERATED {g0}
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('e0', e0()),
            OptionalNamedType('e1', e1()),
            NamedType('e2', e2()),
            DefaultedNamedType('e3', e3('d1'))
        )
        extensionAddition = AdditiveNamedTypes(
            OptionalNamedType('e4', e4()),
            NamedType('e5', e5())
        )
        componentType = rootComponent + extensionAddition
    return SeqOfEnum()


def SCHEMA_sequence_with_extension_addition_groups_1():
    class MySeq(SequenceType):
        '''
        MySeq ::= SEQUENCE {
            s0    OCTET STRING,
            s1    OCTET STRING,
            ...,
            [[
            s2    OCTET STRING,
            s3    OCTET STRING
            ]],
            [[
            s4    OCTET STRING
            ]],
            [[
            s5    OCTET STRING
            ]],
            [[
            s6    OCTET STRING
            ]]
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('s0', OctetStringType()),
            NamedType('s1', OctetStringType())
        )
        extensionAdditionGroups = [
            AdditiveNamedTypes(
                NamedType('s2', OctetStringType()),
                NamedType('s3', OctetStringType())
            ),
            AdditiveNamedTypes(
                NamedType('s4', OctetStringType())
            ),
            AdditiveNamedTypes(
                NamedType('s5', OctetStringType())
            ),
            AdditiveNamedTypes(
                NamedType('s6', OctetStringType())
            )
        ]
        componentType = rootComponent + extensionAdditionGroups
    return MySeq()


def SCHEMA_sequence_with_extension_addition_groups_2():
    class MySeq(SequenceType):
        '''
        MySeq ::= SEQUENCE {
            s0    OCTET STRING,
            ...,
            [[
            s1    OCTET STRING
            ]],
            [[
            s2    OCTET STRING OPTIONAL,
            s3    OCTET STRING OPTIONAL
            ]],
            [[
            s4    OCTET STRING,
            s5    OCTET STRING OPTIONAL,
            s6    OCTET STRING
            ]]
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('s0', OctetStringType())
        )
        extensionAdditionGroups = [
            AdditiveNamedTypes(
                NamedType('s1', OctetStringType())
            ),
            AdditiveNamedTypes(
                OptionalNamedType('s2', OctetStringType()),
                OptionalNamedType('s3', OctetStringType())
            ),
            AdditiveNamedTypes(
                NamedType('s4', OctetStringType()),
                OptionalNamedType('s5', OctetStringType()),
                NamedType('s6', OctetStringType())
            )
        ]
        componentType = rootComponent + extensionAdditionGroups
    return MySeq()


def SCHEMA_sequence_with_extension_addition_groups_3():
    class MySeq(SequenceType):
        '''
        MySeq ::= SEQUENCE {
            s0    OCTET STRING,
            ...,
            s1    OCTET STRING OPTIONAL,
            s2    OCTET STRING,
            [[
            s3    OCTET STRING,
            s4    OCTET STRING,
            s5    OCTET STRING OPTIONAL
            ]],
            [[
            s6    OCTET STRING OPTIONAL
            ]]
        }
        '''
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('s0', OctetStringType())
        )
        extensionAddition = AdditiveNamedTypes(
            OptionalNamedType('s1', OctetStringType()),
            NamedType('s2', OctetStringType())
        )
        extensionAdditionGroups = [
            AdditiveNamedTypes(
                NamedType('s3', OctetStringType()),
                NamedType('s4', OctetStringType()),
                OptionalNamedType('s5', OctetStringType())
            ),
            AdditiveNamedTypes(
                OptionalNamedType('s6', OctetStringType())
            )
        ]
        componentType = rootComponent + extensionAddition + extensionAdditionGroups
    return MySeq()


def DATA_simple_seq(seq, i0_is=False, i0_val=I0_VAL, i1_is=False, i1_val=I1_VAL, i2_is=False, i2_val=I2_VAL,
                    i3_is=False, i3_val=I3_VAL, i4_is=False, i4_val=I4_VAL, ext_mark=False):
    seq = seq(ext_mark)
    if i0_is:
        seq['i0'] = i0(i0_val)
    if i1_is:
        seq['i1'] = i1(i1_val)
    if i2_is:
        seq['i2'] = i2(i2_val)
    if i3_is:
        seq['i3'] = i3(i3_val)
    if i4_is:
        seq['i4'] = i4(i4_val)
    return seq


def DATA_bool_seq(seq, root_val, addition_present, addition_val=False):
    seq = seq()
    seq['b0'] = BooleanType(root_val)
    seq['b1'] = BooleanType(root_val)
    seq['b2'] = BooleanType(root_val)
    seq['b3'] = BooleanType(root_val)
    seq['b4'] = BooleanType(root_val)
    seq['b5'] = BooleanType(root_val)
    seq['b6'] = BooleanType(root_val)
    seq['b7'] = BooleanType(root_val)
    seq['b8'] = BooleanType(root_val)
    seq['b9'] = BooleanType(root_val)
    if addition_present:
        seq['b10'] = BooleanType(addition_val)
        seq['b11'] = BooleanType(addition_val)
        seq['b12'] = BooleanType(addition_val)
        seq['b13'] = BooleanType(addition_val)
        seq['b14'] = BooleanType(addition_val)
        seq['b15'] = BooleanType(addition_val)
        seq['b16'] = BooleanType(addition_val)
        seq['b17'] = BooleanType(addition_val)
        seq['b18'] = BooleanType(addition_val)
        seq['b19'] = BooleanType(addition_val)
        seq['b20'] = BooleanType(addition_val)
    return seq


def DATA_octet_seq(seq, o1_is, o4_is, o5_is):
    seq = seq()
    seq['o0'] = OctetStringType(hexValue='DEAD')
    if o1_is:
        seq['o1'] = OctetStringType(hexValue='BEEF')
    seq['o2'] = o2(hexValue='ABCD')
    seq['o3'] = OctetStringType(hexValue='AA')
    if o4_is:
        seq['o4'] = OctetStringType(hexValue='BBB')
    if o5_is:
        seq['o5'] = OctetStringType(hexValue='CCCC')
    seq['o6'] = OctetStringType(hexValue='DDDDD')
    return seq


def DATA_bitstring_seq(schema, b2_is, b3_is, b9_is):
    seq = schema()
    seq['b0'] = b0(binValue='111100001111')
    seq['b1'] = BitStringType(hexValue='DEAD')
    if b2_is:
        seq['b2'] = BitStringType(hexValue='BEEF')
    if b3_is:
        seq['b3'] = BitStringType(hexValue='ABCD')
    seq['b4'] = b4(binValue='11111')
    seq['b5'] = BitStringType(hexValue='FEED')
    seq['b6'] = b6(binValue='1111000011110000111100001111000011110000')
    seq['b7'] = BitStringType(hexValue='BBBB')
    seq['b8'] = BitStringType(hexValue='CCCCC')
    if b9_is:
        seq['b9'] = b9(binValue='11')
    seq['b10'] = b10(binValue='11110000111100001111')
    seq['b11'] = BitStringType(hexValue='DDDDDD')
    return seq


def DATA_enumerated_seq(schema, e1_is, e3_is, e4_is):
    seq = schema()
    seq['e0'] = 'a1'
    if e1_is:
        seq['e1'] = 'b2'
    seq['e2'] = 'c1'
    if e3_is:
        seq['e3'] = e3('d2')
    if e4_is:
        seq['e4'] = 'f1'
    seq['e5'] = 'g0'
    return seq


def DATA_extension_addition_groups_seq(schema, s0=None, s1=None, s2=None, s3=None, s4=None, s5=None, s6=None):
    seq = schema()
    if s0:
        seq['s0'] = OctetStringType(hexValue=s0)
    if s1:
        seq['s1'] = OctetStringType(hexValue=s1)
    if s2:
        seq['s2'] = OctetStringType(hexValue=s2)
    if s3:
        seq['s3'] = OctetStringType(hexValue=s3)
    if s4:
        seq['s4'] = OctetStringType(hexValue=s4)
    if s5:
        seq['s5'] = OctetStringType(hexValue=s5)
    if s6:
        seq['s6'] = OctetStringType(hexValue=s6)
    return seq


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_simple_seq(SCHEMA_two_elem_seq_with_extension_marker_present_or_not, i0_is=True, i1_is=True, i2_is=False,
                     i3_is=False, i4_is=False, ext_mark=False), '0300870702FCB3'),
    (DATA_simple_seq(SCHEMA_two_elem_seq_with_extension_marker_present_or_not, i0_is=True, i1_is=True, i2_is=False,
                     i3_is=False, i4_is=False, ext_mark=True), '000300870702FCB3'),
    (DATA_simple_seq(SCHEMA_seq_with_only_extension_marker_present_or_not, i0_is=True,i1_is=True, i2_is=True, i3_is=True,
                     i4_is=True, ext_mark=False), '0300870702FCB300178003030D4028'),
    (DATA_simple_seq(SCHEMA_seq_with_only_extension_marker_present_or_not, i0_is=True,i1_is=True, i2_is=True, i3_is=True,
                     i4_is=True, ext_mark=True), '000300870702FCB300178003030D4028'),
])
def test_no_extension_no_default_no_optional_components_present_sequence_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,i1_is=True, i2_is=True, i3_is=True,
                     i4_is=True, ext_mark=True),  '800300870707E00302FCB3020017058003030D400128'),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,i1_is=True, i2_is=True, i3_is=True,
                     i4_is=False, ext_mark=True), '800300870707C00302FCB3020017058003030D40'),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=True, i3_is=False,
                     i4_is=False, ext_mark=True), '800300870707800302FCB3020017'),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=False, i3_is=False,
                     i4_is=False, ext_mark=True), '800300870707000302FCB3'),
])
def test_extension_components_are_present_but_no_default_no_optional_sequence_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True, i1_is=False, i2_is=True, i3_is=False, i4_is=True),
     '0003008707001728'),
    (DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=True),
     '800300870702FCB3001728'),
    (DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True, i1_is=False, i2_is=True, i3_is=True, i4_is=True),
     '400300870700178003030D4028'),
    (DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     'C00300870702FCB300178003030D4028'),
    (DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     'C00300870702FCB3'),
    (DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '8003008707'),
    (DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=False, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '4002FCB3'),
    (DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=False, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '00'),
])
def test_optional_components_present_in_extension_root_sequence_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True, i1_is=False, i2_is=True, i3_is=False, i4_is=True),
     '0003008707001728'),
    (DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=True),
     '800300870702FCB3001728'),
    (DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True, i1_is=False, i2_is=True, i3_is=True, i4_is=True),
     '400300870700178003030D4028'),
    (DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     'C00300870702FCB300178003030D4028'),
])
def test_default_components_with_none_default_values_present_in_extension_root_sequence_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True, i1_val=10, i1_is=True, i2_is=True, i3_val=300, i3_is=False, i4_is=True),
     '0003008707001728'),
    (DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True, i1_val=10, i1_is=False, i2_is=True, i3_val=300, i3_is=True, i4_is=True),
     '0003008707001728'),
    (DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True, i1_val=10, i1_is=True, i2_is=True, i3_val=300, i3_is=True, i4_is=True),
     '0003008707001728'),
])
def test_default_components_with_default_values_present_in_filled_schema_are_not_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False, ext_mark=True),
     '0003008707'),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False, ext_mark=True),
     '800300870707000302FCB3'),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False, ext_mark=True),
     '800300870707800302FCB3020017'),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False, ext_mark=True),
     '800300870707C00302FCB3020017058003030D40'),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True, ext_mark=True),
     '800300870707E00302FCB3020017058003030D400128'),

])
def test_elements_in_extension_addition_may_not_be_present_even_if_no_optional_in_schema_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, exception", [
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=False, i2_is=True, i3_is=False, i4_is=False, ext_mark=True),
     pytest.raises(MandatoryExtensionFieldNotPresent)),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=False, i2_is=False, i3_is=True, i4_is=False, ext_mark=True),
     pytest.raises(MandatoryExtensionFieldNotPresent)),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=True, ext_mark=True),
     pytest.raises(MandatoryExtensionFieldNotPresent)),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=False, i3_is=True, i4_is=False, ext_mark=True),
     pytest.raises(MandatoryExtensionFieldNotPresent)),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=True, ext_mark=True),
     pytest.raises(MandatoryExtensionFieldNotPresent)),
    (DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True, i1_is=False, i2_is=True, i3_is=False, i4_is=True, ext_mark=True),
     pytest.raises(MandatoryExtensionFieldNotPresent)),

])
def test_when_mandatory_extension_field_is_not_present_then_can_not_be_encoded_and_raises(sequence, exception):
    with exception:
        per_encoder(sequence)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=False), '0000'),
    (DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=True, addition_present=False), '7FE0'),
    (DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=True, addition_val=False), '8002BFF801000100010001000100010001000100010001000100'),
    (DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=True, addition_val=True), '8002BFF801800180018001800180018001800180018001800180'),
    (DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=True, addition_present=True, addition_val=False), 'FFE2BFF801000100010001000100010001000100010001000100'),
    (DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=True, addition_present=True, addition_val=True), 'FFE2BFF801800180018001800180018001800180018001800180'),
])
def test_sequence_of_boolean_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=True, o5_is=True), 'E002DEAD02BEEF00ABCD01AA02BBB002CCCC03DDDDD0'),
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=True, o5_is=True), '6002DEAD00ABCD01AA02BBB002CCCC03DDDDD0'),
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=False, o5_is=True), 'A002DEAD02BEEF00ABCD01AA02CCCC03DDDDD0'),
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=True, o5_is=False), 'C002DEAD02BEEF00ABCD01AA02BBB003DDDDD0'),
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=False, o5_is=True), '2002DEAD00ABCD01AA02CCCC03DDDDD0'),
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=True, o5_is=False), '4002DEAD00ABCD01AA02BBB003DDDDD0'),
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=False, o5_is=False), '8002DEAD02BEEF00ABCD01AA03DDDDD0'),
    (DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=False, o5_is=False), '0002DEAD00ABCD01AA03DDDDD0'),
])
def test_sequence_of_octetstring_no_extension_marker_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=True, b9_is=True), 'F00CF0F010DEAD10BEEF10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD'),
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=True, b9_is=True), 'B00CF0F010DEAD10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD'),
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=False, b9_is=True), 'D00CF0F010DEAD10BEEFF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD'),
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=True, b9_is=False), 'F00CF0F010DEAD10BEEF10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD'),
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=False, b9_is=True), '900CF0F010DEADF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD'),
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=True, b9_is=False), 'B00CF0F010DEAD10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD'),
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=False, b9_is=False), 'D00CF0F010DEAD10BEEFF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD'),
    (DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=False, b9_is=False), '900CF0F010DEADF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD'),
])
def test_sequence_of_bitstring_with_extension_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=True, e4_is=True), 'EC0A80E001800100'),
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=False, e3_is=True, e4_is=True), 'AA80E001800100'),
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=False, e4_is=True), 'CC0A0701800100'),
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=True, e4_is=False), 'EC0A80A00100'),
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=False, e3_is=False, e4_is=True), '8A0701800100'),
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=False, e3_is=True, e4_is=False), 'AA80A00100'),
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=False, e4_is=False), 'CC0A050100'),
    (DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=False, e3_is=False, e4_is=False), '8A050100'),
])
def test_sequence_of_enumerated_with_extension_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence, encoded", [
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_1,
                                        s0='99', s1='AA'), '00019901AA'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_1,
                                        s0='99', s1='AA', s2='BB', s3='CC'), '80019901AA07000401BB01CC'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_1,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD'), '80019901AA07800401BB01CC0201DD'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_1,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE'), '80019901AA07C00401BB01CC0201DD0201EE'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_1,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE', s6='FF'), '80019901AA07E00401BB01CC0201DD0201EE0201FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99'), '000199'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA'), '80019905000201AA'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA', s2='BB'), '80019905800201AA038001BB'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA', s3='CC'), '80019905800201AA034001CC'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA', s2='BB', s3='CC'), '80019905800201AA05C001BB01CC'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE', s6='FF'), '80019905C00201AA05C001BB01CC078001DD01EE01FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s6='FF'), '80019905C00201AA05C001BB01CC050001DD01FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA', s4='DD', s5='EE', s6='FF'), '80019905400201AA078001DD01EE01FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_2,
                                        s0='99', s1='AA', s4='DD', s6='FF'), '80019905400201AA050001DD01FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99'), '000199'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s1='AA'), '80019907000201AA'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s1='AA', s2='BB'), '80019907800201AA0201BB'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s2='BB'), '80019906800201BB'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD'), '80019907C00201AA0201BB050001CC01DD'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE'), '80019907C00201AA0201BB078001CC01DD01EE'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE', s6='FF'), '80019907E00201AA0201BB078001CC01DD01EE038001FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s6='FF'), '80019907E00201AA0201BB050001CC01DD038001FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s2='BB', s3='CC', s4='DD', s6='FF'), '80019906E00201BB050001CC01DD038001FF'),
    (DATA_extension_addition_groups_seq(SCHEMA_sequence_with_extension_addition_groups_3,
                                        s0='99', s2='BB', s3='CC', s4='DD'), '80019906C00201BB050001CC01DD'),
])
def test_sequence_with_extension_addition_groups_can_be_encoded(sequence, encoded):
    assert per_encoder(sequence) == bytearray.fromhex(encoded)


# elementy w extensionAddition musza byc obecne po kolei
# jesli pierwszy element nie jest OPTIONAL, to musi on byc obecny jesli chcecmy zakodowac
# element drugi itd
# Jesli element pierwszy jesli OPTIONAL to nie musi on byc obecny jesli chcemy zakodowac
# element drugi


from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class MyInt1(IntegerType):
    subtypeSpec = ValueRange(-5, -1)


class MyOctetString1(OctetStringType):
    subtypeSpec = ValueSize(1, 4)


class MyBitString1(BitStringType):
    pass


class MyBoolean1(BooleanType):
    pass


class MyEnumerated1(EnumeratedType):
    enumerationRoot = NamedValues(
        ('four', 0),
        ('five', 1),
        ('six', 2),
    )
    namedValues = enumerationRoot


MyInt2 = MyInt1


MyInt3 = MyInt2


MyOctetString2 = MyOctetString1


MyOctetString3 = MyOctetString2


MyBitString2 = MyBitString1


MyBitString3 = MyBitString2


MyBoolean2 = MyBoolean1


MyBoolean3 = MyBoolean2


MyEnumerated2 = MyEnumerated1


MyEnumerated3 = MyEnumerated2


class MySeq1(SequenceType):
    rootComponent = AdditiveNamedTypes(
        DefaultedNamedType('d00', IntegerType(10)),
        DefaultedNamedType('d01', IntegerType(0)),
    )
    componentType = rootComponent


MySeq4 = MySeq1


MySeq2 = MySeq1


class MySeq3(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('d03', MyInt2()),
        NamedType('d04', MyInt3()),
        NamedType('d05', MyOctetString2()),
        NamedType('d06', MyOctetString3()),
        NamedType('d07', MyBitString2()),
        NamedType('d08', MyBitString3()),
        NamedType('d09', MyBoolean2()),
        NamedType('d10', MyBoolean3()),
        NamedType('d11', MyEnumerated2()),
        NamedType('d12', MyEnumerated3()),
    )
    componentType = rootComponent


MySeq5 = MySeq3





from ..type import DefType, SimpleType
from asn1PERser.classes.module import typereference_to_type
from asn1PERser.classes.templates.creator import template_filler


class DefinedType(DefType):
    def __init__(self, typereference):
        super(DefinedType, self).__init__()
        self.typereference = typereference

    def fill_template(self):
        defined_type = typereference_to_type[self.typereference]

        if self.identifier and self.typereference and not self.Value:  # Foo ::= Bar
            return template_filler.fill(asn_type='type_type', type_or_value='type',
                                        identifier=self.identifier.replace('-', '_'),
                                        typereference=self.typereference.replace('-', '_'))

        if self.Value and self.valuereference and self.typereference:  # foo SomeType ::= value
            while not defined_type.__class__.__name__ in ['IntegerType', 'BooleanType', 'OctetStringType', 'BitStringType']:
                defined_type = typereference_to_type[defined_type.typereference]
            defined_type.valuereference = self.valuereference
            defined_type.typereference = self.typereference
            defined_type.Value = self.Value
            return template_filler.fill(asn_type='simple_value', type_or_value='value',
                                        valuereference=self.valuereference,
                                        class_type=self.typereference.replace('-','_'),
                                        value=defined_type.Value)
        if self.Value:
            defined_type.Value = self.Value
            return defined_type.fill_template()
        if defined_type.__class__.__name__ != self.__class__.__name__:
            return defined_type.fill_template()
        return template_filler.fill(asn_type=self.__class__.__name__, type_or_value='type',
                                    class_name=self.template_field_type, class_type=defined_type.template_class_name)
